@@ -247,12 +247,9 @@ gdk_input_device_new (GdkDisplay  *display,
  error:
 
   g_free (gdkdev->info.name);
-  if (gdkdev->axes)
-    g_free (gdkdev->axes);
-  if (gdkdev->info.keys)
-    g_free (gdkdev->info.keys);
-  if (gdkdev->info.axes)
-    g_free (gdkdev->info.axes);
+  g_free (gdkdev->axes);
+  g_free (gdkdev->info.keys);
+  g_free (gdkdev->info.axes);
   g_object_unref (gdkdev);
   
   return NULL;
