@@ -48,6 +48,7 @@
 #include "gtkfilechooserdialog.h"
 #include "gtkfilechooserprivate.h"
 #include "gtkfilechooserutils.h"
+#include "gtkmarshalers.h"
 
 #include "gtkfilechooserbutton.h"
 
@@ -86,6 +87,13 @@ enum
   PROP_WIDTH_CHARS
 };
 
+/* Signals */
+enum
+{
+  FILE_SET,
+  LAST_SIGNAL
+};
+
 /* TreeModel Columns */
 enum
 {
@@ -219,7 +227,7 @@ static void     gtk_file_chooser_button_drag_data_received (GtkWidget        *wi
 							    gint              x,
 							    gint              y,
 							    GtkSelectionData *data,
-							    guint             info,
+							    guint             type,
 							    guint             drag_time);
 static void     gtk_file_chooser_button_show_all           (GtkWidget        *widget);
 static void     gtk_file_chooser_button_hide_all           (GtkWidget        *widget);
@@ -301,6 +309,7 @@ static void     dialog_response_cb               (GtkDialog      *dialog,
 						  gint            response,
 						  gpointer        user_data);
 
+static guint file_chooser_button_signals[LAST_SIGNAL] = { 0 };
 
 /* ******************* *
  *  GType Declaration  *
@@ -342,6 +351,26 @@ gtk_file_chooser_button_class_init (GtkFileChooserButtonClass * class)
   widget_class->style_set = gtk_file_chooser_button_style_set;
   widget_class->screen_changed = gtk_file_chooser_button_screen_changed;
   widget_class->mnemonic_activate = gtk_file_chooser_button_mnemonic_activate;
+  
+  /**
+   * GtkFileChooserButtons::file-set:
+   * @widget: the object which received the signal.
+   *
+   * The ::file-set signal is emitted when the user selects a file.
+   * 
+   * Note that this signal is only emitted when the <emphasis>user</emphasis>
+   * changes the file. 
+   *
+   * Since: 2.12
+   */
+  file_chooser_button_signals[FILE_SET] =
+    g_signal_new (I_("file-set"),
+		  G_TYPE_FROM_CLASS (gobject_class),
+		  G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
+		  G_STRUCT_OFFSET (GtkFileChooserButtonClass, file_set),
+		  NULL, NULL,
+		  _gtk_marshal_VOID__VOID,
+		  G_TYPE_NONE, 0);
 
   /**
    * GtkFileChooserButton:dialog:
@@ -1033,7 +1062,7 @@ gtk_file_chooser_button_drag_data_received (GtkWidget	     *widget,
 					    gint	      x,
 					    gint	      y,
 					    GtkSelectionData *data,
-					    guint	      info,
+					    guint	      type,
 					    guint	      drag_time)
 {
   GtkFileChooserButton *button = GTK_FILE_CHOOSER_BUTTON (widget);
@@ -1045,13 +1074,13 @@ gtk_file_chooser_button_drag_data_received (GtkWidget	     *widget,
     (*GTK_WIDGET_CLASS (gtk_file_chooser_button_parent_class)->drag_data_received) (widget,
 										    context,
 										    x, y,
-										    data, info,
+										    data, type,
 										    drag_time);
 
   if (widget == NULL || context == NULL || data == NULL || data->length < 0)
     return;
 
-  switch (info)
+  switch (type)
     {
     case TEXT_URI_LIST:
       {
@@ -1593,6 +1622,7 @@ model_add_special_get_info_cb (GtkFileSystemHandle *handle,
   GdkPixbuf *pixbuf;
   GtkFileSystemHandle *model_handle;
   struct ChangeIconThemeData *data = user_data;
+  const gchar *name;
 
   if (!data->button->priv->model)
     /* button got destroyed */
@@ -1630,9 +1660,13 @@ model_add_special_get_info_cb (GtkFileSystemHandle *handle,
       g_object_unref (pixbuf);
     }
 
-  gtk_list_store_set (GTK_LIST_STORE (data->button->priv->model), &iter,
-		      DISPLAY_NAME_COLUMN, gtk_file_info_get_display_name (info),
-		      -1);
+  gtk_tree_model_get (data->button->priv->model, &iter,
+                      DISPLAY_NAME_COLUMN, &name,
+                      -1);
+  if (!name)
+    gtk_list_store_set (GTK_LIST_STORE (data->button->priv->model), &iter,
+  		        DISPLAY_NAME_COLUMN, gtk_file_info_get_display_name (info),
+		        -1);
 
 out:
   g_object_unref (data->button);
@@ -1688,15 +1722,9 @@ model_add_special (GtkFileChooserButton *button)
 			  -1);
 
       button->priv->n_special++;
-
-#ifndef G_OS_WIN32
-      desktopdir = g_build_filename (homedir, DESKTOP_DISPLAY_NAME, NULL);
-#endif
     }
 
-#ifdef G_OS_WIN32
-  desktopdir = _gtk_file_system_win32_get_desktop ();
-#endif
+  desktopdir = g_get_user_special_dir (G_USER_DIRECTORY_DESKTOP);
 
   if (desktopdir)
     {
@@ -1705,7 +1733,6 @@ model_add_special (GtkFileChooserButton *button)
       struct ChangeIconThemeData *info;
 
       path = gtk_file_system_filename_to_path (button->priv->fs, desktopdir);
-      g_free (desktopdir);
       gtk_list_store_insert (store, &iter, pos);
       pos++;
 
@@ -2213,7 +2240,7 @@ update_combo_box (GtkFileChooserButton *button)
             if (base_path)
               {
 	        row_found = (paths &&
-		  	     paths->data &&
+			     paths->data &&
 			     gtk_file_path_compare (base_path, paths->data) == 0);
 	        gtk_file_path_free (base_path);
               }
@@ -2698,6 +2725,8 @@ dialog_response_cb (GtkDialog *dialog,
 
   gtk_widget_set_sensitive (priv->combo_box, TRUE);
   gtk_widget_hide (priv->dialog);
+
+  g_signal_emit_by_name (user_data, "file-set");
 }
 
 
