@@ -57,6 +57,7 @@ enum {
   ACTIVATE_CURRENT,
   CANCEL,
   CYCLE_FOCUS,
+  MOVE_SELECTED,
   LAST_SIGNAL
 };
 
@@ -193,6 +194,8 @@ static void gtk_real_menu_shell_cycle_focus      (GtkMenuShell      *menu_shell,
 static void     gtk_menu_shell_reset_key_hash    (GtkMenuShell *menu_shell);
 static gboolean gtk_menu_shell_activate_mnemonic (GtkMenuShell *menu_shell,
 						  GdkEventKey  *event);
+static gboolean gtk_menu_shell_real_move_selected (GtkMenuShell  *menu_shell, 
+						   gint           distance);
 
 static guint menu_shell_signals[LAST_SIGNAL] = { 0 };
 
@@ -237,6 +240,7 @@ gtk_menu_shell_class_init (GtkMenuShellClass *klass)
   klass->cancel = gtk_real_menu_shell_cancel;
   klass->select_item = gtk_menu_shell_real_select_item;
   klass->insert = gtk_menu_shell_real_insert;
+  klass->move_selected = gtk_menu_shell_real_move_selected;
 
   menu_shell_signals[DEACTIVATE] =
     g_signal_new (I_("deactivate"),
@@ -289,7 +293,27 @@ gtk_menu_shell_class_init (GtkMenuShellClass *klass)
 			     _gtk_marshal_VOID__ENUM,
 			     G_TYPE_NONE, 1,
 			     GTK_TYPE_DIRECTION_TYPE);
-
+  /**
+   * GtkMenuShell::move-selected:
+   * @menu_shell: the object on which the signal is emitted
+   * @distance: +1 to move to the next item, -1 to move to the previous
+   *
+   * The ::move-selected signal is emitted to move the selection to
+   * another item. 
+   * 
+   * Returns: %TRUE to stop the signal emission, %FALSE to continue
+   *
+   * Since: 2.12
+   */ 
+  menu_shell_signals[MOVE_SELECTED] =
+    g_signal_new (I_("move_selected"),
+		  G_OBJECT_CLASS_TYPE (object_class),
+		  G_SIGNAL_RUN_LAST,
+		  G_STRUCT_OFFSET (GtkMenuShellClass, move_selected),
+		  _gtk_boolean_handled_accumulator, NULL,
+		  _gtk_marshal_BOOLEAN__INT,
+		  G_TYPE_BOOLEAN, 1,
+		  G_TYPE_INT);
 
   binding_set = gtk_binding_set_by_class (klass);
   gtk_binding_entry_add_signal (binding_set,
@@ -789,18 +813,17 @@ gtk_menu_shell_enter_notify (GtkWidget        *widget,
 	  (GTK_IS_MENU_ITEM (menu_item) && 
 	   !_gtk_menu_item_is_selectable (menu_item)))
 	return TRUE;
-      
-      if ((menu_item->parent == widget) &&
-	  (menu_shell->active_menu_item != menu_item) &&
+
+      if (menu_item->parent == widget &&
 	  GTK_IS_MENU_ITEM (menu_item))
 	{
 	  if (menu_shell->ignore_enter)
 	    return TRUE;
-	  
-	  if ((event->detail != GDK_NOTIFY_INFERIOR) &&
-	      (GTK_WIDGET_STATE (menu_item) != GTK_STATE_PRELIGHT))
-	    {
-	      gtk_menu_shell_select_item (menu_shell, menu_item);
+
+	  if (event->detail != GDK_NOTIFY_INFERIOR)
+            {
+	      if (GTK_WIDGET_STATE (menu_item) != GTK_STATE_PRELIGHT)
+                gtk_menu_shell_select_item (menu_shell, menu_item);
 
               /* If any mouse button is down, and there is a submenu
                * that is not yet visible, activate it. It's sufficient
@@ -1133,9 +1156,9 @@ gtk_menu_shell_activate_item (GtkMenuShell      *menu_shell,
 }
 
 /* Distance should be +/- 1 */
-static void
-gtk_menu_shell_move_selected (GtkMenuShell  *menu_shell, 
-			      gint           distance)
+static gboolean
+gtk_menu_shell_real_move_selected (GtkMenuShell  *menu_shell, 
+				   gint           distance)
 {
   if (menu_shell->active_menu_item)
     {
@@ -1186,6 +1209,19 @@ gtk_menu_shell_move_selected (GtkMenuShell  *menu_shell,
       if (node)
 	gtk_menu_shell_select_item (menu_shell, node->data);
     }
+
+  return TRUE;
+}
+
+/* Distance should be +/- 1 */
+static void
+gtk_menu_shell_move_selected (GtkMenuShell  *menu_shell, 
+			      gint           distance)
+{
+  gboolean handled = FALSE;
+
+  g_signal_emit (menu_shell, menu_shell_signals[MOVE_SELECTED], 0,
+		 distance, &handled);
 }
 
 /**
@@ -1262,6 +1298,9 @@ gtk_menu_shell_select_submenu_first (GtkMenuShell     *menu_shell)
 {
   GtkMenuItem *menu_item;
 
+  if (menu_shell->active_menu_item == NULL)
+    return FALSE;
+
   menu_item = GTK_MENU_ITEM (menu_shell->active_menu_item); 
   
   if (menu_item->submenu)
