@@ -470,7 +470,7 @@ win32_poll_status_timeout (GtkPrintOperation *op)
   win32_poll_status (op);
 
   if (!gtk_print_operation_is_finished (op))
-    op_win32->timeout_id = g_timeout_add (STATUS_POLLING_TIME,
+    op_win32->timeout_id = gdk_threads_add_timeout (STATUS_POLLING_TIME,
 					  (GSourceFunc)win32_poll_status_timeout,
 					  op);
   g_object_unref (op);
@@ -511,7 +511,7 @@ win32_end_run (GtkPrintOperation *op,
     {
       op_win32->printerHandle = printerHandle;
       win32_poll_status (op);
-      op_win32->timeout_id = g_timeout_add (STATUS_POLLING_TIME,
+      op_win32->timeout_id = gdk_threads_add_timeout (STATUS_POLLING_TIME,
 					    (GSourceFunc)win32_poll_status_timeout,
 					    op);
     }
@@ -1356,7 +1356,7 @@ create_application_page (GtkPrintOperation *op)
   WORD baseunitX, baseunitY;
   WORD *array;
   GtkRequisition requisition;
-  const char *app_name;
+  const char *tab_label;
 
   /* Make the template the size of the custom widget size request */
   gtk_widget_size_request (op->priv->custom_widget, &requisition);
@@ -1386,10 +1386,13 @@ create_application_page (GtkPrintOperation *op)
   page.dwFlags = PSP_DLGINDIRECT | PSP_USETITLE | PSP_PREMATURE;
   page.hInstance = GetModuleHandle (NULL);
   page.pResource = template;
-  app_name = g_get_application_name ();
-  if (app_name == NULL)
-    app_name = _("Application");
-  page.pszTitle = g_utf8_to_utf16 (app_name, 
+  
+  tab_label = op->priv->custom_tab_label;
+  if (tab_label == NULL)
+    tab_label = g_get_application_name ();
+  if (tab_label == NULL)
+    tab_label = _("Application");
+  page.pszTitle = g_utf8_to_utf16 (tab_label, 
 				   -1, NULL, NULL, NULL);
   page.pfnDlgProc = pageDlgProc;
   page.pfnCallback = NULL;
