@@ -296,7 +296,7 @@ gtk_recent_manager_init (GtkRecentManager *manager)
   priv->filename = g_build_filename (g_get_home_dir (),
 				     GTK_RECENTLY_USED_FILE,
 				     NULL);
-  priv->poll_timeout = g_timeout_add (POLL_DELTA,
+  priv->poll_timeout = gdk_threads_add_timeout (POLL_DELTA,
 		  		      gtk_recent_manager_poll_timeout,
 				      manager);
 
@@ -360,8 +360,7 @@ gtk_recent_manager_finalize (GObject *object)
   if (priv->poll_timeout)
     g_source_remove (priv->poll_timeout);
   
-  if (priv->filename)
-    g_free (priv->filename);
+  g_free (priv->filename);
   
   if (priv->recent_items)
     g_bookmark_file_free (priv->recent_items);
@@ -402,6 +401,7 @@ gtk_recent_manager_real_changed (GtkRecentManager *manager)
       g_bookmark_file_to_file (priv->recent_items,
 		               priv->filename,
 			       &write_error);
+
       if (write_error)
         {
           filename_warning ("Attempting to store changes into `%s', "
@@ -411,6 +411,8 @@ gtk_recent_manager_real_changed (GtkRecentManager *manager)
 	  g_error_free (write_error);
 	}
 
+      priv->write_in_progress = FALSE;
+	  
       /* we have sync'ed our list with the storage file, so we
        * update the file mtime in order to skip the timed check
        * and spare us from a re-read.
@@ -422,8 +424,6 @@ gtk_recent_manager_real_changed (GtkRecentManager *manager)
 			    priv->filename,
 			    g_strerror (errno));
 
-	  priv->write_in_progress = FALSE;
-	  
 	  g_object_thaw_notify (G_OBJECT (manager));
 
 	  return;
@@ -507,7 +507,7 @@ gtk_recent_manager_set_filename (GtkRecentManager *manager,
     }
 
   priv->filename = g_strdup (filename);
-  priv->poll_timeout = g_timeout_add (POLL_DELTA,
+  priv->poll_timeout = gdk_threads_add_timeout (POLL_DELTA,
 		  		      gtk_recent_manager_poll_timeout,
 				      manager);
 
@@ -613,7 +613,7 @@ build_recent_items_list (GtkRecentManager *manager)
  * monitors the recently used resources list, and emits the "changed" signal
  * each time something inside the list changes.
  *
- * #GtkRecentManager objects are expansive: be sure to create them only when
+ * #GtkRecentManager objects are expensive: be sure to create them only when
  * needed. You should use the gtk_recent_manager_new_for_screen() or the
  * gtk_recent_manager_get_default() functions instead.
  *
@@ -634,9 +634,9 @@ gtk_recent_manager_new (void)
  * gtk_recent_manager_get_for_screen().
  *
  * Return value: A unique #GtkRecentManager associated with the
- *   default screen. This recent manager is associated to the
+ *   default screen. This recent manager is associated with the
  *   screen and can be used as long as the screen is open.
- *   Do no ref or unref it.
+ *   Do not ref or unref it.
  *
  * Since: 2.10
  */
@@ -835,18 +835,16 @@ gboolean
 gtk_recent_manager_add_item (GtkRecentManager  *manager,
 			     const gchar       *uri)
 {
-  GtkRecentData *recent_data;
-  GError *add_error;
+  GtkRecentData recent_data;
   gboolean retval;
   
   g_return_val_if_fail (GTK_IS_RECENT_MANAGER (manager), FALSE);
   g_return_val_if_fail (uri != NULL, FALSE);
 
-  recent_data = g_slice_new (GtkRecentData);
-  
-  recent_data->display_name = NULL;
-  recent_data->description = NULL;
-  
+  recent_data.display_name = NULL;
+  recent_data.description = NULL;
+  recent_data.mime_type = NULL;
+
 #ifdef G_OS_UNIX
   if (has_case_prefix (uri, "file:/"))
     {
@@ -854,34 +852,33 @@ gtk_recent_manager_add_item (GtkRecentManager  *manager,
       const gchar *mime_type;
       
       filename = g_filename_from_uri (uri, NULL, NULL);
-      mime_type = xdg_mime_get_mime_type_for_file (filename, NULL);
-      if (!mime_type)
-        recent_data->mime_type = g_strdup (GTK_RECENT_DEFAULT_MIME);
-      else
-        recent_data->mime_type = g_strdup (mime_type);
+      if (filename)
+        {
+          mime_type = xdg_mime_get_mime_type_for_file (filename, NULL);
+          if (mime_type)
+            recent_data.mime_type = g_strdup (mime_type);
       
-      g_free (filename);
+          g_free (filename);
+        }
+
+      if (!recent_data.mime_type)
+        recent_data.mime_type = g_strdup (GTK_RECENT_DEFAULT_MIME);
     }
   else
 #endif
-    recent_data->mime_type = g_strdup (GTK_RECENT_DEFAULT_MIME);
+    recent_data.mime_type = g_strdup (GTK_RECENT_DEFAULT_MIME);
   
-  recent_data->app_name = g_strdup (g_get_application_name ());
-  recent_data->app_exec = g_strjoin (" ", g_get_prgname (), "%u", NULL);
+  recent_data.app_name = g_strdup (g_get_application_name ());
+  recent_data.app_exec = g_strjoin (" ", g_get_prgname (), "%u", NULL);
+  recent_data.groups = NULL;
+  recent_data.is_private = FALSE;
   
-  recent_data->groups = NULL;
+  retval = gtk_recent_manager_add_full (manager, uri, &recent_data);
   
-  recent_data->is_private = FALSE;
-  
-  add_error = NULL;
-  retval = gtk_recent_manager_add_full (manager, uri, recent_data);
-  
-  g_free (recent_data->mime_type);
-  g_free (recent_data->app_name);
-  g_free (recent_data->app_exec);
+  g_free (recent_data.mime_type);
+  g_free (recent_data.app_name);
+  g_free (recent_data.app_exec);
 
-  g_slice_free (GtkRecentData, recent_data);
-  
   return retval;
 }
 
@@ -1234,7 +1231,7 @@ gtk_recent_manager_lookup_item (GtkRecentManager  *manager,
       return NULL;
     }
  
-  return gtk_recent_info_ref (info);
+  return info;
 }
 
 /**
@@ -1321,6 +1318,9 @@ gtk_recent_manager_get_items (GtkRecentManager *manager)
   priv = manager->priv;
   if (!priv->recent_items)
     return NULL;
+
+  if (priv->limit == 0)
+    return NULL;
   
   uris = g_bookmark_file_get_uris (priv->recent_items, &uris_len);
   for (i = 0; i < uris_len; i++)
@@ -1438,7 +1438,7 @@ gtk_recent_info_get_type (void)
   static GType info_type = 0;
   
   if (!info_type)
-    info_type = g_boxed_type_register_static ("GtkRecentInfo",
+    info_type = g_boxed_type_register_static (I_("GtkRecentInfo"),
     					      (GBoxedCopyFunc) gtk_recent_info_ref,
     					      (GBoxedFreeFunc) gtk_recent_info_unref);
   return info_type;
@@ -1733,11 +1733,9 @@ recent_app_info_free (RecentAppInfo *app_info)
   if (!app_info)
     return;
   
-  if (app_info->name)
-    g_free (app_info->name);
+  g_free (app_info->name);
   
-  if (app_info->exec)
-    g_free (app_info->exec);
+  g_free (app_info->exec);
   
   g_free (app_info);
 }
@@ -2051,9 +2049,14 @@ gtk_recent_info_get_icon (GtkRecentInfo *info,
   if (info->mime_type)
     retval = get_icon_for_mime_type (info->mime_type, size);
 
-  /* this should never fail */  
+  /* this function should never fail */  
   if (!retval)
-    retval = get_icon_fallback (GTK_STOCK_FILE, size);
+    {
+      if (info->mime_type && strcmp (info->mime_type, "x-directory/normal") == 0)
+        retval = get_icon_fallback (GTK_STOCK_DIRECTORY, size);
+      else
+        retval = get_icon_fallback (GTK_STOCK_FILE, size);
+    }
   
   return retval;
 }
@@ -2221,7 +2224,7 @@ get_uri_shortname_for_display (const gchar *uri)
       
       local_file = g_filename_from_uri (uri, NULL, NULL);
       
-      if (local_file != NULL)
+      if (local_file)
         {
           name = g_filename_display_basename (local_file);
           validated = TRUE;
@@ -2229,7 +2232,8 @@ get_uri_shortname_for_display (const gchar *uri)
 		
       g_free (local_file);
     } 
-  else
+  
+  if (!name)
     {
       gchar *method;
       gchar *local_file;
@@ -2291,7 +2295,9 @@ gtk_recent_info_get_short_name (GtkRecentInfo *info)
  * gtk_recent_info_get_uri_display:
  * @info: a #GtkRecentInfo
  *
- * Gets a displayable version of the resource's URI.
+ * Gets a displayable version of the resource's URI.  If the resource
+ * is local, it returns a local path; if the resource is not local,
+ * it returns the UTF-8 encoded content of gtk_recent_info_get_uri().
  *
  * Return value: a UTF-8 string containing the resource's URI or %NULL
  *
@@ -2300,18 +2306,28 @@ gtk_recent_info_get_short_name (GtkRecentInfo *info)
 gchar *
 gtk_recent_info_get_uri_display (GtkRecentInfo *info)
 {
-  gchar *filename, *filename_utf8;
+  gchar *retval;
   
   g_return_val_if_fail (info != NULL, NULL);
-  
-  filename = g_filename_from_uri (info->uri, NULL, NULL);
-  if (!filename)
-    return NULL;
-      
-  filename_utf8 = g_filename_to_utf8 (filename, -1, NULL, NULL, NULL);
-  g_free (filename);
 
-  return filename_utf8;
+  retval = NULL;
+  if (gtk_recent_info_is_local (info))
+    {
+      gchar *filename;
+
+      filename = g_filename_from_uri (info->uri, NULL, NULL);
+      if (!filename)
+        return NULL;
+      
+      retval = g_filename_to_utf8 (filename, -1, NULL, NULL, NULL);
+      g_free (filename);
+    }
+  else
+    {
+      retval = make_valid_utf8 (info->uri);
+    }
+
+  return retval;
 }
 
 /**
